"""
    nbkode.dop853
    ~~~~~~~~~~~~~

    Methods of Runge-Kutta family:
    - DOP853

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from typing import Callable

import numpy as np

from nbkode import corevs, dop853_coefficients
from nbkode.corevs import MAX_FACTOR, MIN_FACTOR, SAFETY, rk_step
from nbkode.nbcompat import clip, numba


def step_builder(A, B, C, E5, E3, error_exponent):
    """Perform a single fixed step.

    This outer function should only contains attributes
    associated with the solver class not with the solver instance.

    Parameters
    ----------
    A : ndarray, shape (n_stages, )
        Coefficients for combining previous stages to compute the next
        stage.

    Returns
    -------
    callable

    """

    @numba.njit
    def _step(t_bound, rhs, t, y, f, h, K, rtol, atol, max_step):
        """Perform a single fixed step.

        This inner function should only contains attributes
        associated with the solver instance not with the solver class.

        Parameters
        ----------
        t_bound : float
            Boundary time - the integration won’t continue beyond it.
            It also determines the direction of the integration.
        rhs : callable
            Right-hand side of the system.
        ts : ndarray, shape (LEN_HISTORY, )
            Last ORDER + 1 times
        ys : ndarray, shape (LEN_HISTORY, n)
            Last ORDER + 1 values, for each n dimension
        fs : ndarray, shape (LEN_HISTORY, n)
            Last ORDER + 1 value of the derivative, for each n dimension
        h : ndarray, shape (1, )
            Step to use.
        K : ndarray, shape (n_stages + 1, n)
            Storage array for putting RK stages here. Stages are stored in rows.
            The last row is a linear combination of the previous rows with
            coefficients
        rtol : float
            Tolerance (relative) for termination.
        atol : float
            The allowable error of the zero value. If `func` is complex-valued,
            a larger `tol` is recommended as both the real and imaginary parts
            of `x` contribute to ``|x - x0|``.
        max_step : float
            Maximum allowed step size.

        ts, ts and fs are modified in place.
        h and K is modified in place.

        Returns
        -------
        int
            number of steps given (0 or 1)
        """

        t_cur = t[-1]
        y_cur = y[-1]
        f_cur = f[-1]

        min_step = 10 * (np.nextafter(t_cur, np.inf) - t_cur)
        _h = clip(h[0], min_step, max_step)
        step_rejected = False
        while True:
            if _h < min_step:
                raise RuntimeError("Required step is too small.")

            t_new = min(t_cur + _h, t_bound)
            _h = t_new - t_cur

            if _h == 0:
                return 0

            y_new, f_new = rk_step(rhs, t_cur, y_cur, f_cur, _h, A, B, C, K)

            # Estimate norm of scaled error
            scale = atol + np.maximum(np.abs(y_cur), np.abs(y_new)) * rtol

            err5 = np.dot(K.T, E5) / scale
            err3 = np.dot(K.T, E3) / scale
            err5_norm_2 = np.linalg.norm(err5) ** 2
            err3_norm_2 = np.linalg.norm(err3) ** 2
            if err5_norm_2 == 0 and err3_norm_2 == 0:
                error_norm = 0.0
            else:
                denom = err5_norm_2 + 0.01 * err3_norm_2
                error_norm = np.abs(_h) * err5_norm_2 / np.sqrt(denom * len(scale))

            if error_norm < 1:
                if error_norm == 0:
                    factor = MAX_FACTOR
                else:
                    factor = min(MAX_FACTOR, SAFETY * error_norm ** error_exponent)

                if step_rejected:
                    factor = min(1, factor)

                _h *= factor

                break
            else:
                _h *= max(MIN_FACTOR, SAFETY * error_norm ** error_exponent)
                step_rejected = True

        f[:-1] = f[1:]
        f[-1] = f_new

        t[:-1] = t[1:]
        t[-1] = t_new

        y[:-1] = y[1:]
        y[-1] = y_new

        h[0] = _h

        return 1

    return _step


class DOP853(corevs.VariableStepRungeKutta):
    """Explicit Runge-Kutta method of order 8.

    This is a Python implementation of "DOP853" algorithm originally written
    in Fortran [#]_, [#]_. Note that this is not a literate translation, but
    the algorithmic core and coefficients are the same.
    Can be applied in the complex domain.


    References
    ----------
    .. [#] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
           Equations I: Nonstiff Problems", Sec. II.
    .. [#] `Page with original Fortran code of DOP853
            <http://www.unige.ch/~hairer/software.html>`_.
    """

    GROUP = "Runge-Kutta"
    IMPLICIT = False

    n_stages = dop853_coefficients.N_STAGES
    order = 8
    error_estimator_order = 7
    A = np.ascontiguousarray(dop853_coefficients.A[:n_stages, :n_stages])
    B = dop853_coefficients.B
    C = np.ascontiguousarray(dop853_coefficients.C[:n_stages])
    E3 = dop853_coefficients.E3
    E5 = dop853_coefficients.E5
    D = dop853_coefficients.D

    A_EXTRA = np.ascontiguousarray(dop853_coefficients.A[n_stages + 1 :])
    C_EXTRA = np.ascontiguousarray(dop853_coefficients.C[n_stages + 1 :])

    _step_builder = step_builder

    @classmethod
    def _step_builder_args(cls):
        return (cls.A, cls.B, cls.C, cls.E5, cls.E3, cls.error_exponent)

    def __init__(
        self,
        rhs: Callable,
        t0: float,
        y0: np.ndarray,
        params: np.ndarray = None,
        *,
        max_step=np.inf,
        rtol=1e-3,
        atol=1e-6,
    ):
        super().__init__(rhs, t0, y0, params, max_step=max_step, rtol=rtol, atol=atol)

        self.K_extended = np.empty(
            (dop853_coefficients.N_STAGES_EXTENDED, y0.size), dtype=y0.dtype
        )
        self.K = self.K_extended[: self.n_stages + 1]



"""
    nbkode.corefs
    ~~~~~~~~~~~~~

    Definitions for linear multistep methods.

    See: https://en.wikipedia.org/wiki/Linear_multistep_method

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from typing import Callable

import numpy as np

from nbkode.nbcompat import numba

from .core import Solver
from .nbcompat import j_newton, newton_hd


def forward_step_builder(A):
    """Perform a single fixed step.

    This outer function should only contains attributes
    associated with the solver class not with the solver instance.

    Parameters
    ----------
    A : ndarray, shape (1, ORDER + 1)
        Coefficients for combining previous stages to compute the next
        stage.

    Returns
    -------
    callable
    """

    @numba.njit
    def _step(t_bound, rhs, ts, ys, fs, h):
        """Perform a single fixed step.

        This inner function should only contains attributes
        associated with the solver instance not with the solver class.

        Parameters
        ----------
        t_bound : float
            Boundary time - the integration won’t continue beyond it.
            It also determines the direction of the integration.
        rhs : Callable
            Right-hand side of the system.
        ts : ndarray, shape (LEN_HISTORY, )
            Last ORDER + 1 times
        ys : ndarray, shape (LEN_HISTORY, n)
            Last ORDER + 1 values, for each n dimension
        fs : ndarray, shape (LEN_HISTORY, n)
            Last ORDER + 1 value of the derivative, for each n dimension
        h : float
            Step to use.

        ts, ys and fs are modified in place.

        Returns
        -------
        int
            number of steps given (0 or 1)
        """

        t_new = ts[-1] + h
        if t_new > t_bound:
            return 0
        f_new = rhs(ts[-1], ys[-1])

        fs[:-1] = fs[1:]
        fs[-1] = f_new

        y_new = ys[-1] + h * A @ fs

        ts[:-1] = ts[1:]
        ts[-1] = t_new

        ys[:-1] = ys[1:]
        ys[-1] = y_new

        return 1

    return _step


def backward_step_builder(A):
    """Perform a single fixed step.

    This outer function should only contains attributes
    associated with the solver class not with the solver instance.

    Parameters
    ----------
    A : ndarray, shape (n_stages, )
        Coefficients for combining previous stages to compute the next
        stage.

    Returns
    -------
    callable

    """

    @numba.njit
    def _to_solve(y_new, rhs, t_new, y_last, f, h):
        f[-1] = rhs(t_new, y_last)

        return y_last + h * (A @ f).ravel() - y_new

    @numba.njit
    def _step(t_bound, rhs, ts, ys, fs, h, rtol, atol, max_iter):
        """Perform a single fixed step.

        This inner function should only contains attributes
        associated with the solver instance not with the solver class.

        Parameters
        ----------
        t_bound : float
            Boundary time - the integration won’t continue beyond it.
            It also determines the direction of the integration.
        rhs : Callable
            Right-hand side of the system.
        ts : ndarray, shape (LEN_HISTORY, )
            Last ORDER + 1 times
        ys : ndarray, shape (LEN_HISTORY, n)
            Last ORDER + 1 values, for each n dimension
        fs : ndarray, shape (LEN_HISTORY, n)
            Last ORDER + 1 value of the derivative, for each n dimension
        h : float
            Step to use.
        rtol : float, optional
            Tolerance (relative) for termination.
        atol : float, optional
            The allowable error of the zero value. If `func` is complex-valued,
            a larger `tol` is recommended as both the real and imaginary parts
            of `x` contribute to ``|x - x0|``.
        max_iter : int, optional
            Maximum number of iterations.

        ts, ys and fs are modified in place.

        Returns
        -------
        int
            number of steps given (0 or 1)
        """

        t_new = ts[-1] + h
        if t_new > t_bound:
            return 0

        fs[:-1] = fs[1:]

        # TODO: For size > 1, we need something like this.
        # y_new = root(_to_solve, ys[-1], args=(rhs, t_new, ys[-1], fs)).x
        if ys[-1].size == 1:
            y_new = j_newton(
                _to_solve,
                ys[-1],
                args=(rhs, t_new, ys[-1], fs, h),
                tol=atol,
                rtol=rtol,
                maxiter=max_iter,
            )
        else:
            y_new = newton_hd(
                _to_solve,
                ys[-1],
                args=(rhs, t_new, ys[-1], fs, h),
                atol=atol,
                rtol=rtol,
                maxiter=max_iter,
            )
        ts[:-1] = ts[1:]
        ts[-1] = t_new

        ys[:-1] = ys[1:]
        ys[-1] = y_new

        return 1

    return _step


class _FixedStepBaseSolver(Solver):
    """Forward fixed step solver.

    Parameters
    ----------
    h : float
        Step size
    first_stepper_cls : Solver or 'str' or None
        - subclass of Solver.
        - 'auto'
        - str that will be used as getattr(nbkode, first_stepper)
        - None

    See `Solver` for information about the other arguments.
    """

    COEFS: np.ndarray

    ORDER: int

    FIXED_STEP = True

    FIRST_STEPPER_CLS = "RungeKutta45"

    @classmethod
    def _step_builder_args(cls):
        return (cls.COEFS,)

    def __init__(
        self,
        rhs: Callable,
        t0: float,
        y0: np.ndarray,
        params: np.ndarray = None,
        *,
        h: float = 1,
        first_stepper_cls=None,
    ):
        super().__init__(rhs, t0, y0, params)

        # TODO: CHECK VALID VALUES
        self._h = h

        # Code to handle the first steps in multistep methods
        # Some of these methods require to build up a history of states
        # integrated using different methods.
        if first_stepper_cls == "auto":
            first_stepper_cls = self.FIRST_STEPPER_CLS

        if first_stepper_cls is not None:
            if isinstance(first_stepper_cls, str):
                import nbkode

                first_stepper_cls = getattr(nbkode, first_stepper_cls)

            if first_stepper_cls.FIXED_STEP:
                # For fixed step solver we do N steps with the step size.
                solver = first_stepper_cls(rhs, t0, y0, params, h=h)
                for ndx in range(self.ORDER - 1):
                    solver.step()
                    self.push(solver.t, solver.y, solver.f)
            else:
                # For variable step solver we run N times until h, 2h, 3h .. (ORDER - 1)h
                solver = first_stepper_cls(rhs, t0, y0, params)
                for ndx in range(1, self.ORDER):
                    solver.run(ndx * h)
                    self.push(solver.t, solver.y, solver.f)

    def __init_subclass__(cls, **kwargs):
        if hasattr(cls, "COEFS"):
            cls.COEFS = np.ascontiguousarray(cls.COEFS, dtype=float).flatten()
            cls.ORDER = cls.COEFS.size
            if cls.ORDER == 1:
                cls.COEFS = np.append(cls.COEFS, 0.0)
            cls.COEFS.shape = (1, cls.COEFS.size)
            cls.LEN_HISTORY = cls.COEFS.size

        super().__init_subclass__(**kwargs)

    @staticmethod
    @numba.njit()
    def _interpolate(t_eval, rhs, ts, ys, fs, *extra_args):
        """Interpolate solution at t_eval"""

        if not (ts[0] <= t_eval <= ts[-1]):
            raise ValueError("Time to interpolate outside range")

        y_out = np.empty(ys[0].shape)
        for ndx in range(len(y_out)):
            y_out[ndx] = np.interp(t_eval, ts, ys[:, ndx])

        return y_out

    @property
    def step_size(self):
        return self._h

    def push(self, t_new, y_new, f_new):
        self._ts[:-1] = self._ts[1:]
        self._ts[-1] = t_new

        self._ys[:-1] = self._ys[1:]
        self._ys[-1] = y_new

        self._fs[:-1] = self._fs[1:]
        self._fs[-1] = f_new


class FFixedStepBaseSolver(_FixedStepBaseSolver):
    """Forward fixed step solver.

    Parameters
    ----------
    h : float
        Step size

    See `Solver` for information about the other arguments.
    """

    IMPLICIT = False

    _step_builder = forward_step_builder

    def __init__(
        self,
        rhs: Callable,
        t0: float,
        y0: np.ndarray,
        params: np.ndarray = None,
        *,
        h: float = 1,
        first_stepper_cls=None,
    ):
        super().__init__(rhs, t0, y0, params, h=h, first_stepper_cls=first_stepper_cls)

    def _step_extra_args(self):
        return (self._h,)


class BFixedStepBaseSolver(_FixedStepBaseSolver):
    """Backward fixed step solver.

    Parameters
    ----------
    h : float
        Step size
    rtol : float, optional
        Tolerance (relative) for termination.
    atol : float, optional
        The allowable error of the zero value. If `func` is complex-valued,
        a larger `tol` is recommended as both the real and imaginary parts
        of `x` contribute to ``|x - x0|``.
    max_iter : int, optional
        Maximum number of iterations.

    `rtol`, `atol` and `max_iter` are given to solve the implicit equation.

    See `Solver` for information about the other arguments.
    """

    IMPLICIT = True

    _step_builder = backward_step_builder

    def __init__(
        self,
        rhs: Callable,
        t0: float,
        y0: np.ndarray,
        params: np.ndarray = None,
        *,
        h: float = 1,
        rtol=0.0,
        atol=1.48e-8,
        max_iter=50,
        first_stepper_cls=None,
    ):
        super().__init__(rhs, t0, y0, params, h=h, first_stepper_cls=first_stepper_cls)

        self.atol = atol
        self.rtol = rtol
        self.max_iter = max_iter

    def _step_extra_args(self):
        return self._h, self.rtol, self.atol, self.max_iter



"""
    nbkode.testsuite.solver_fs
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test fixed step solvers for:
    - argument handling
    - jittability
    - running different methods.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np
import pytest

import nbkode

solvers_i = nbkode.get_solvers(implicit=True, fixed_step=True)
solvers_e = nbkode.get_solvers(implicit=False, fixed_step=True)
solvers = nbkode.get_solvers(fixed_step=True)


y0_1 = np.atleast_1d(1.0)
y0_2 = np.atleast_1d([1.0, 2.0])


def f1(t, x, k):
    return -k * x


def f2(t, x, k):
    return -k * x


@pytest.mark.parametrize("solver", solvers)
def test_f1_step(solver):
    solver = solver(f1, 0.0, y0_1, params=(0.01,))
    solver.step()
    solver.run(1)


@pytest.mark.parametrize("solver", solvers)
def test_f1(solver):

    solver = solver(f1, 0.0, y0_1, params=(0.01,))
    solver.run(10)

    # TODO: This is a rather large tolerance.
    np.testing.assert_allclose(solver.y, np.exp(-0.01 * 10), atol=0.1)


@pytest.mark.parametrize("solver", solvers)
def test_f2_step(solver):
    solver = solver(f2, 0.0, y0_2, params=(0.01, 0.05))
    solver.step()
    solver.run(1)


@pytest.mark.parametrize("solver", solvers)
def test_f2(solver):

    solver = solver(f2, 0.0, y0_2, params=(0.01, 0.05))
    solver.run(10)

    # TODO: This is a rather large tolerance.
    np.testing.assert_allclose(solver.y[0], np.exp(-0.01 * 10), rtol=0.15)
    np.testing.assert_allclose(solver.y[1], 2.0 * np.exp(-0.05 * 10), rtol=0.15)


@pytest.mark.parametrize("solver", solvers)
def test_first_stepper(solver):
    sol = solver(f1, 0.0, y0_1, params=(0.01,), first_stepper_cls=None)
    assert sol.t == 0.0
    sol.step()
    assert sol.t == sol.step_size

    sol = solver(f1, 0.0, y0_1, params=(0.01,), first_stepper_cls="Euler")
    assert sol.t == (sol.ORDER - 1) * sol.step_size

    sol = solver(f1, 0.0, y0_1, params=(0.01,), first_stepper_cls=nbkode.Euler)
    assert sol.t == (sol.ORDER - 1) * sol.step_size

    sol = solver(f1, 0.0, y0_1, params=(0.01,), first_stepper_cls="RungeKutta45")
    assert sol.t == (sol.ORDER - 1) * sol.step_size

    sol = solver(f1, 0.0, y0_1, params=(0.01,), first_stepper_cls=nbkode.RungeKutta45)
    assert sol.t == (sol.ORDER - 1) * sol.step_size

    sol2 = solver(f1, 0.0, y0_1, params=(0.01,), first_stepper_cls="auto")
    assert sol.t == sol2.t



#!/usr/bin/env python3
#
# numbakit-ode documentation build configuration file, created by
# sphinx-quickstart on Thu Mar  1 13:33:14 2012.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import datetime

try:
    from importlib.metadata import version
except ImportError:
    # Backport for Python < 3.8
    from importlib_metadata import version

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.mathjax",
    "matplotlib.sphinxext.plot_directive",
    "nbsphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "numbakit-ode"
author = "Hernan E. Grecco"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

try:  # pragma: no cover
    version = version(project)
except Exception:  # pragma: no cover
    # we seem to have a local copy not installed without setuptools
    # so the reported version will be unknown
    version = "unknown"

release = version
this_year = datetime.date.today().year
copyright = "%s, %s" % (this_year, author)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme = 'default'
html_theme = "flask"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []
html_theme_path = ["_themes"]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}
html_sidebars = {
    "index": ["sidebarintro.html", "sourcelink.html", "searchbox.html"],
    "**": [
        "sidebarlogo.html",
        "localtoc.html",
        "relations.html",
        "sourcelink.html",
        "searchbox.html",
    ],
}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "numbakit-odedoc"


# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    "preamble": "".join((r"\DeclareUnicodeCharacter{2212}{-}",))  # MINUS
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "numbakit-ode.tex",
        "numbakit-ode Documentation",
        "Hernan E. Grecco",
        "manual",
    )
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", "numbakit-ode", "numbakit-ode Documentation", ["Hernan E. Grecco"], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "numbakit-ode",
        "numbakit-ode Documentation",
        "Hernan E. Grecco",
        "numbakit-ode",
        "One line description of project.",
        "Miscellaneous",
    )
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'


# -- Options for Epub output ---------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The language of the text. It defaults to the language option
# or en if the language is not set.
# epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
# epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
# epub_identifier = ''

# A unique identification for the text.
# epub_uid = ''

# A tuple containing the cover image and cover page html template filenames.
# epub_cover = ()

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
# epub_post_files = []

# A list of files that should not be packed into the epub file.
# epub_exclude_files = []

# The depth of the table of contents in toc.ncx.
# epub_tocdepth = 3

# Allow duplicate toc entries.
# epub_tocdup = True


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {"http://docs.python.org/": None}

# -- Doctest configuration -----------------------------------------------------

# fill a dictionary with package names that may be missing
# this is checked by :skipif: clauses in certain doctests that rely
# on optional dependencies
doctest_global_setup = """
from importlib.util import find_spec

not_installed = {
    pkg_name: find_spec(pkg_name) is None
    for pkg_name in [
        'uncertainties',
        'serialize',
    ]
}
"""



"""
    nbkode.testsuite.test_nbcompat
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test for non-trivial implementation of python/numpy/scipy functions
    which are not njitable (yet).

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


import numpy as np
import pytest
from numpy.testing import (
    assert_,
    assert_allclose,
    assert_array_equal,
    assert_equal,
    assert_warns,
    suppress_warnings,
)
from scipy.optimize import zeros as sp_zeros

from nbkode.nbcompat import numba, zeros

cos = np.cos
exp = np.exp
sin = np.sin
power = np.power
newton = zeros.newton


# A few test functions used frequently:
# # A simple quadratic, (x-1)^2 - 1
def f1(x):
    return x ** 2 - 2 * x - 1


def f1_1(x):
    return 2 * x - 2


def f1_2(x):
    return 2.0 + 0 * x


def f1_and_p_and_pp(x):
    return f1(x), f1_1(x), f1_2(x)


# Simple transcendental function
def f2(x):
    return exp(x) - cos(x)


def f2_1(x):
    return exp(x) + sin(x)


def f2_2(x):
    return exp(x) + cos(x)


class TestNewton:
    @staticmethod
    def f1(x):
        return x ** 2 - 2 * x - 1  # == (x-1)**2 - 2

    @staticmethod
    def f1_1(x):
        return 2 * x - 2

    @staticmethod
    def f1_2(x):
        return 2.0 + 0 * x

    @staticmethod
    def f2(x):
        return exp(x) - cos(x)

    @staticmethod
    def f2_1(x):
        return exp(x) + sin(x)

    @staticmethod
    def f2_2(x):
        return exp(x) + cos(x)

    def test_newton(self):
        for f, f_1, f_2 in [
            (self.f1, self.f1_1, self.f1_2),
            (self.f2, self.f2_1, self.f2_2),
        ]:
            x = zeros.newton(f, 3, tol=1e-6)
            assert_allclose(f(x), 0, atol=1e-6)
            x = zeros.newton(f, 3, x1=5, tol=1e-6)  # secant, x0 and x1
            assert_allclose(f(x), 0, atol=1e-6)
            x = zeros.newton(f, 3, fprime=f_1, tol=1e-6)  # newton
            assert_allclose(f(x), 0, atol=1e-6)
            x = zeros.newton(f, 3, fprime=f_1, fprime2=f_2, tol=1e-6)  # halley
            assert_allclose(f(x), 0, atol=1e-6)

    def test_array_newton(self):
        """test newton with array"""

        def f1(x, *a):
            b = a[0] + x * a[3]
            return a[1] - a[2] * (np.exp(b / a[5]) - 1.0) - b / a[4] - x

        def f1_1(x, *a):
            b = a[3] / a[5]
            return -a[2] * np.exp(a[0] / a[5] + x * b) * b - a[3] / a[4] - 1

        def f1_2(x, *a):
            b = a[3] / a[5]
            return -a[2] * np.exp(a[0] / a[5] + x * b) * b ** 2

        a0 = np.array(
            [
                5.32725221,
                5.48673747,
                5.49539973,
                5.36387202,
                4.80237316,
                1.43764452,
                5.23063958,
                5.46094772,
                5.50512718,
                5.42046290,
            ]
        )
        a1 = (np.sin(range(10)) + 1.0) * 7.0
        args = (a0, a1, 1e-09, 0.004, 10, 0.27456)
        x0 = [7.0] * 10
        x = zeros.newton(f1, x0, f1_1, args)
        x_expected = (
            6.17264965,
            11.7702805,
            12.2219954,
            7.11017681,
            1.18151293,
            0.143707955,
            4.31928228,
            10.5419107,
            12.7552490,
            8.91225749,
        )
        assert_allclose(x, x_expected)
        # test halley's
        x = zeros.newton(f1, x0, f1_1, args, fprime2=f1_2)
        assert_allclose(x, x_expected)
        # test secant
        x = zeros.newton(f1, x0, args=args)
        assert_allclose(x, x_expected)

    def test_array_newton_complex(self):
        def f(x):
            return x + 1 + 1j

        def fprime(x):
            return 1.0

        t = np.full(4, 1j)
        x = zeros.newton(f, t, fprime=fprime)
        assert_allclose(f(x), 0.0)

        # should work even if x0 is not complex
        t = np.ones(4)
        x = zeros.newton(f, t, fprime=fprime)
        assert_allclose(f(x), 0.0)

        x = zeros.newton(f, t)
        assert_allclose(f(x), 0.0)

    def test_array_secant_active_zero_der(self):
        """test secant doesn't continue to iterate zero derivatives"""
        x = zeros.newton(
            lambda x, *a: x * x - a[0], x0=[4.123, 5], args=[np.array([17, 25])]
        )
        assert_allclose(x, (4.123105625617661, 5.0))

    def test_array_newton_integers(self):
        # test secant with float
        x = zeros.newton(lambda y, z: z - y ** 2, [4.0] * 2, args=([15.0, 17.0],))
        assert_allclose(x, (3.872983346207417, 4.123105625617661))
        # test integer becomes float
        x = zeros.newton(lambda y, z: z - y ** 2, [4] * 2, args=([15, 17],))
        assert_allclose(x, (3.872983346207417, 4.123105625617661))

    def test_array_newton_zero_der_failures(self):
        # test derivative zero warning
        assert_warns(
            RuntimeWarning,
            zeros.newton,
            lambda y: y ** 2 - 2,
            [0.0, 0.0],
            lambda y: 2 * y,
        )
        # test failures and zero_der
        with pytest.warns(RuntimeWarning):
            results = zeros.newton(
                lambda y: y ** 2 - 2, [0.0, 0.0], lambda y: 2 * y, full_output=True
            )
            assert_allclose(results.root, 0)
            assert results.zero_der.all()
            assert not results.converged.any()

    def test_newton_full_output(self):
        # Test the full_output capability, both when converging and not.
        # Use simple polynomials, to avoid hitting platform dependencies
        # (e.g., exp & trig) in number of iterations

        x0 = 3
        expected_counts = [(6, 7), (5, 10), (3, 9)]

        for derivs in range(3):
            kwargs = {
                "tol": 1e-6,
                "full_output": True,
            }
            for k, v in [["fprime", self.f1_1], ["fprime2", self.f1_2]][:derivs]:
                kwargs[k] = v

            x, r = zeros.newton(self.f1, x0, disp=False, **kwargs)
            assert_(r.converged)
            assert_equal(x, r.root)
            assert_equal((r.iterations, r.function_calls), expected_counts[derivs])
            if derivs == 0:
                assert r.function_calls <= r.iterations + 1
            else:
                assert_equal(r.function_calls, (derivs + 1) * r.iterations)

            # Now repeat, allowing one fewer iteration to force convergence failure
            iters = r.iterations - 1
            x, r = zeros.newton(self.f1, x0, maxiter=iters, disp=False, **kwargs)
            assert_(not r.converged)
            assert_equal(x, r.root)
            assert_equal(r.iterations, iters)

            if derivs == 1:
                # Check that the correct Exception is raised and
                # validate the start of the message.
                with pytest.raises(
                    RuntimeError,
                    match="Failed to converge after %d iterations, value is .*"
                    % (iters),
                ):
                    x, r = zeros.newton(self.f1, x0, maxiter=iters, disp=True, **kwargs)

    def test_deriv_zero_warning(self):
        func = lambda x: x ** 2 - 2.0
        dfunc = lambda x: 2 * x
        assert_warns(RuntimeWarning, zeros.newton, func, 0.0, dfunc, disp=False)
        with pytest.raises(RuntimeError, match="Derivative was zero"):
            zeros.newton(func, 0.0, dfunc)

    def test_newton_does_not_modify_x0(self):
        # https://github.com/scipy/scipy/issues/9964
        x0 = np.array([0.1, 3])
        x0_copy = x0.copy()  # Copy to test for equality.
        newton(np.sin, x0, np.cos)
        assert_array_equal(x0, x0_copy)

    def test_maxiter_int_check(self):
        for method in [
            zeros.newton,
        ]:
            with pytest.raises(
                TypeError, match="'float' object cannot be interpreted as an integer"
            ):
                method(f1, 0.0, 1.0, maxiter=72.45)


def test_complex_halley():
    """Test Halley's works with complex roots"""

    def f(x, *a):
        return a[0] * x ** 2 + a[1] * x + a[2]

    def f_1(x, *a):
        return 2 * a[0] * x + a[1]

    def f_2(x, *a):
        retval = 2 * a[0]
        try:
            size = len(x)
        except TypeError:
            return retval
        else:
            return [retval] * size

    z = complex(1.0, 2.0)
    coeffs = (2.0, 3.0, 4.0)
    y = zeros.newton(f, z, args=coeffs, fprime=f_1, fprime2=f_2, tol=1e-6)
    # (-0.75000000000000078+1.1989578808281789j)
    assert_allclose(f(y, *coeffs), 0, atol=1e-6)
    z = [z] * 10
    coeffs = (2.0, 3.0, 4.0)
    y = zeros.newton(f, z, args=coeffs, fprime=f_1, fprime2=f_2, tol=1e-6)
    assert_allclose(f(y, *coeffs), 0, atol=1e-6)


def test_zero_der_nz_dp():
    """Test secant method with a non-zero dp, but an infinite newton step"""
    # pick a symmetrical functions and choose a point on the side that with dx
    # makes a secant that is a flat line with zero slope, EG: f = (x - 100)**2,
    # which has a root at x = 100 and is symmetrical around the line x = 100
    # we have to pick a really big number so that it is consistently true
    # now find a point on each side so that the secant has a zero slope
    dx = np.finfo(float).eps ** 0.33
    # 100 - p0 = p1 - 100 = p0 * (1 + dx) + dx - 100
    # -> 200 = p0 * (2 + dx) + dx
    p0 = (200.0 - dx) / (2.0 + dx)
    with suppress_warnings() as sup:
        sup.filter(RuntimeWarning, "RMS of")
        x = zeros.newton(lambda y: (y - 100.0) ** 2, x0=[p0] * 10)
    assert_allclose(x, [100] * 10)
    # test scalar cases too
    p0 = (2.0 - 1e-4) / (2.0 + 1e-4)
    with suppress_warnings() as sup:
        sup.filter(RuntimeWarning, "Tolerance of")
        x = zeros.newton(lambda y: (y - 1.0) ** 2, x0=p0, disp=False)
    assert_allclose(x, 1)
    with pytest.raises(RuntimeError, match="Tolerance of"):
        x = zeros.newton(lambda y: (y - 1.0) ** 2, x0=p0, disp=True)
    p0 = (-2.0 + 1e-4) / (2.0 + 1e-4)
    with suppress_warnings() as sup:
        sup.filter(RuntimeWarning, "Tolerance of")
        x = zeros.newton(lambda y: (y + 1.0) ** 2, x0=p0, disp=False)
    assert_allclose(x, -1)
    with pytest.raises(RuntimeError, match="Tolerance of"):
        x = zeros.newton(lambda y: (y + 1.0) ** 2, x0=p0, disp=True)


def test_array_newton_failures():
    """Test that array newton fails as expected"""
    # p = 0.68  # [MPa]
    # dp = -0.068 * 1e6  # [Pa]
    # T = 323  # [K]
    diameter = 0.10  # [m]
    # L = 100  # [m]
    roughness = 0.00015  # [m]
    rho = 988.1  # [kg/m**3]
    mu = 5.4790e-04  # [Pa*s]
    u = 2.488  # [m/s]
    reynolds_number = rho * u * diameter / mu  # Reynolds number

    def colebrook_eqn(darcy_friction, re, dia):
        return 1 / np.sqrt(darcy_friction) + 2 * np.log10(
            roughness / 3.7 / dia + 2.51 / re / np.sqrt(darcy_friction)
        )

    # only some failures
    with pytest.warns(RuntimeWarning):
        result = zeros.newton(
            colebrook_eqn,
            x0=[0.01, 0.2, 0.02223, 0.3],
            maxiter=2,
            args=[reynolds_number, diameter],
            full_output=True,
        )
        assert not result.converged.all()
    # they all fail
    with pytest.raises(RuntimeError):
        result = zeros.newton(
            colebrook_eqn,
            x0=[0.01] * 2,
            maxiter=2,
            args=[reynolds_number, diameter],
            full_output=True,
        )


# this test should **not** raise a RuntimeWarning
def test_gh8904_zeroder_at_root_fails():
    """Test that Newton or Halley don't warn if zero derivative at root"""

    # a function that has a zero derivative at it's root
    def f_zeroder_root(x):
        return x ** 3 - x ** 2

    # should work with secant
    r = zeros.newton(f_zeroder_root, x0=0)
    assert_allclose(r, 0, atol=sp_zeros._xtol, rtol=sp_zeros._rtol)
    # test again with array
    r = zeros.newton(f_zeroder_root, x0=[0] * 10)
    assert_allclose(r, 0, atol=sp_zeros._xtol, rtol=sp_zeros._rtol)

    # 1st derivative
    def fder(x):
        return 3 * x ** 2 - 2 * x

    # 2nd derivative
    def fder2(x):
        return 6 * x - 2

    # should work with newton and halley
    r = zeros.newton(f_zeroder_root, x0=0, fprime=fder)
    assert_allclose(r, 0, atol=sp_zeros._xtol, rtol=sp_zeros._rtol)
    r = zeros.newton(f_zeroder_root, x0=0, fprime=fder, fprime2=fder2)
    assert_allclose(r, 0, atol=sp_zeros._xtol, rtol=sp_zeros._rtol)
    # test again with array
    r = zeros.newton(f_zeroder_root, x0=[0] * 10, fprime=fder)
    assert_allclose(r, 0, atol=sp_zeros._xtol, rtol=sp_zeros._rtol)
    r = zeros.newton(f_zeroder_root, x0=[0] * 10, fprime=fder, fprime2=fder2)
    assert_allclose(r, 0, atol=sp_zeros._xtol, rtol=sp_zeros._rtol)

    # also test that if a root is found we do not raise RuntimeWarning even if
    # the derivative is zero, EG: at x = 0.5, then fval = -0.125 and
    # fder = -0.25 so the next guess is 0.5 - (-0.125/-0.5) = 0 which is the
    # root, but if the solver continued with that guess, then it will calculate
    # a zero derivative, so it should return the root w/o RuntimeWarning
    r = zeros.newton(f_zeroder_root, x0=0.5, fprime=fder)
    assert_allclose(r, 0, atol=sp_zeros._xtol, rtol=sp_zeros._rtol)
    # test again with array
    r = zeros.newton(f_zeroder_root, x0=[0.5] * 10, fprime=fder)
    assert_allclose(r, 0, atol=sp_zeros._xtol, rtol=sp_zeros._rtol)
    # doesn't apply to halley


def test_gh_8881():
    r"""Test that Halley's method realizes that the 2nd order adjustment
    is too big and drops off to the 1st order adjustment."""
    n = 9

    def f(x):
        return power(x, 1.0 / n) - power(n, 1.0 / n)

    def fp(x):
        return power(x, (1.0 - n) / n) / n

    def fpp(x):
        return power(x, (1.0 - 2 * n) / n) * (1.0 / n) * (1.0 - n) / n

    x0 = 0.1
    # The root is at x=9.
    # The function has positive slope, x0 < root.
    # Newton succeeds in 8 iterations
    rt, r = newton(f, x0, fprime=fp, full_output=True)
    assert r.converged
    # Before the Issue 8881/PR 8882, halley would send x in the wrong direction.
    # Check that it now succeeds.
    rt, r = newton(f, x0, fprime=fp, fprime2=fpp, full_output=True)
    assert r.converged


def test_gh_9608_preserve_array_shape():
    """
    Test that shape is preserved for array inputs even if fprime or fprime2 is
    scalar
    """

    def f(x):
        return x ** 2

    def fp(x):
        return 2 * x

    def fpp(x):
        return 2

    x0 = np.array([-2], dtype=np.float32)
    rt, r = newton(f, x0, fprime=fp, fprime2=fpp, full_output=True)
    assert r.converged

    x0_array = np.array([-2, -3], dtype=np.float32)
    # This next invocation should fail
    with pytest.raises(IndexError):
        result = zeros.newton(f, x0_array, fprime=fp, fprime2=fpp, full_output=True)

    def fpp_array(x):
        return np.full(np.shape(x), 2, dtype=np.float32)

    result = zeros.newton(f, x0_array, fprime=fp, fprime2=fpp_array, full_output=True)
    assert result.converged.all()


def test_gh9551_raise_error_if_disp_true():
    """Test that if disp is true then zero derivative raises RuntimeError"""

    def f(x):
        return x * x + 1

    def f_p(x):
        return 2 * x

    assert_warns(RuntimeWarning, zeros.newton, f, 1.0, f_p, disp=False)
    with pytest.raises(
        RuntimeError,
        match=r"^Derivative was zero\. Failed to converge after \d+ iterations, value is [+-]?\d*\.\d+\.$",
    ):
        zeros.newton(f, 1.0, f_p)
    root = zeros.newton(f, complex(10.0, 10.0), f_p)
    assert_allclose(root, complex(0.0, 1.0))


def test_jitted():
    f1j = numba.njit()(f1)
    assert_allclose(zeros.j_newton(f1j, 0), zeros.newton(f1, 0))


from scipy.optimize import root


def test_ndnewton():
    @numba.njit()
    def fff(x):
        return x ** 2.0 - 10.0

    fff(0.0)
    y0 = np.asarray([1.0, 1.0, 1.0])
    assert_allclose(root(fff, y0).x, zeros.newton_hd(fff, y0))



"""
    nbkode.runge_kutta
    ~~~~~~~~~~~~~~~~~~

    Methods of Runge-Kutta family:
    - RungeKutta23
    - RungeKutta45

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np

from . import corevs
from .corevs import MAX_FACTOR, MIN_FACTOR, SAFETY, rk_step
from .nbcompat import clip, numba  # noqa: F401


def step_builder(A, B, C, E, error_exponent):
    """Perform a single fixed step.

    This outer function should only contains attributes
    associated with the solver class not with the solver instance.

    Parameters
    ----------
    A : ndarray, shape (n_stages, n_stages)
        Coefficients for combining previous RK stages to compute the next
        stage. For explicit methods the coefficients at and above the main
        diagonal are zeros.
    B : ndarray, shape (n_stages,)
        Coefficients for combining RK stages for computing the final
        prediction.
    C : ndarray, shape (n_stages,)
        Coefficients for incrementing time for consecutive RK stages.
        The value for the first stage is always zero.
    E :
    error_exponent :

    Returns
    -------
    callable
    """

    @numba.njit
    def _step(t_bound, rhs, t, y, f, h, K, rtol, atol, max_step):
        """Perform a single fixed step.

        This inner function should only contains attributes
        associated with the solver instance not with the solver class.

        Parameters
        ----------
        t_bound : float
            Boundary time - the integration won’t continue beyond it.
            It also determines the direction of the integration.
        rhs : callable
            Right-hand side of the system.
        ts : ndarray, shape (LEN_HISTORY, )
            Last ORDER + 1 times
        ys : ndarray, shape (LEN_HISTORY, n)
            Last ORDER + 1 values, for each n dimension
        fs : ndarray, shape (LEN_HISTORY, n)
            Last ORDER + 1 value of the derivative, for each n dimension
        h : float
            Step to use.
        K : ndarray, shape (n_stages + 1, n)
            Storage array for putting RK stages here. Stages are stored in rows.
            The last row is a linear combination of the previous rows with
            coefficients
        rtol : float
            Tolerance (relative) for termination.
        atol : float
            The allowable error of the zero value. If `func` is complex-valued,
            a larger `tol` is recommended as both the real and imaginary parts
            of `x` contribute to ``|x - x0|``.
        max_step : float
            Maximum allowed step size.

        ts, ts and fs are modified in place.
        h and K is modified in place.

        Returns
        -------
        int
            number of steps given (0 or 1)
        """

        t_cur = t[-1]
        y_cur = y[-1]
        f_cur = f[-1]

        min_step = 10 * (np.nextafter(t_cur, np.inf) - t_cur)
        _h = clip(h[0], min_step, max_step)
        step_rejected = False
        while True:
            if _h < min_step:
                raise RuntimeError("Required step is too small.")

            t_new = min(t_cur + _h, t_bound)
            _h = t_new - t_cur

            if _h == 0:
                return 0

            y_new, f_new = rk_step(rhs, t_cur, y_cur, f_cur, _h, A, B, C, K)

            # Estimate norm of scaled error
            scale = atol + np.maximum(np.abs(y_cur), np.abs(y_new)) * rtol
            scaled_error = (K.T @ E) * _h / scale
            error_norm = (np.sum(scaled_error ** 2) / scaled_error.size) ** 0.5

            if error_norm < 1:
                if error_norm == 0:
                    factor = MAX_FACTOR
                else:
                    factor = min(MAX_FACTOR, SAFETY * error_norm ** error_exponent)

                if step_rejected:
                    factor = min(1, factor)

                _h *= factor

                break
            else:
                _h *= max(MIN_FACTOR, SAFETY * error_norm ** error_exponent)
                step_rejected = True

        f[:-1] = f[1:]
        f[-1] = f_new

        t[:-1] = t[1:]
        t[-1] = t_new

        y[:-1] = y[1:]
        y[-1] = y_new

        h[0] = _h

        return 1

    return _step


class RungeKutta23(corevs.VariableStepRungeKutta):
    """Explicit Runge-Kutta method of order 3(2).

    This uses the Bogacki-Shampine pair of formulas [#]_. The error is controlled
    assuming accuracy of the second-order method, but steps are taken using the
    third-order accurate formula (local extrapolation is done). A cubic Hermite
    polynomial is used for the dense output.
    Can be applied in the complex domain.



    References
    ----------
    .. [#] P. Bogacki, L.F. Shampine, "A 3(2) Pair of Runge-Kutta Formulas",
       Appl. Math. Lett. Vol. 2, No. 4. pp. 321-325, 1989.
    """

    GROUP = "Runge-Kutta"
    IMPLICIT = False

    order = 3
    error_estimator_order = 2
    n_stages = 3
    C = np.ascontiguousarray([0, 1 / 2, 3 / 4], dtype=float)
    A = np.ascontiguousarray([[0, 0, 0], [1 / 2, 0, 0], [0, 3 / 4, 0]], dtype=float)
    B = np.ascontiguousarray([2 / 9, 1 / 3, 4 / 9], dtype=float)
    E = np.ascontiguousarray([5 / 72, -1 / 12, -1 / 9, 1 / 8], dtype=float)
    P = np.ascontiguousarray(
        [[1, -4 / 3, 5 / 9], [0, 1, -2 / 3], [0, 4 / 3, -8 / 9], [0, -1, 1]],
        dtype=float,
    )

    _step_builder = step_builder

    @classmethod
    def _step_builder_args(cls):
        return cls.A, cls.B, cls.C, cls.E, cls.error_exponent


class RungeKutta45(corevs.VariableStepRungeKutta):
    """Explicit Runge-Kutta method of order 5(4).

    This uses the Dormand-Prince pair of formulas [#]_. The error is controlled
    assuming accuracy of the fourth-order method accuracy, but steps are taken
    using the fifth-order accurate formula (local extrapolation is done).
    A quartic interpolation polynomial is used for the dense output [#]_.
    Can be applied in the complex domain.

    References
    ----------
    .. [#] J. R. Dormand, P. J. Prince, "A family of embedded Runge-Kutta
           formulae", Journal of Computational and Applied Mathematics, Vol. 6,
           No. 1, pp. 19-26, 1980.
    .. [#] L. W. Shampine, "Some Practical Runge-Kutta Formulas", Mathematics
           of Computation,, Vol. 46, No. 173, pp. 135-150, 1986.
    """

    GROUP = "Runge-Kutta"
    IMPLICIT = False

    order = 5
    error_estimator_order = 4
    n_stages = 6
    C = np.ascontiguousarray([0, 1 / 5, 3 / 10, 4 / 5, 8 / 9, 1], dtype=float)
    A = np.ascontiguousarray(
        [
            [0, 0, 0, 0, 0],
            [1 / 5, 0, 0, 0, 0],
            [3 / 40, 9 / 40, 0, 0, 0],
            [44 / 45, -56 / 15, 32 / 9, 0, 0],
            [19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729, 0],
            [9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656],
        ],
        dtype=float,
    )
    B = np.ascontiguousarray(
        [35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84], dtype=float
    )
    E = np.ascontiguousarray(
        [-71 / 57600, 0, 71 / 16695, -71 / 1920, 17253 / 339200, -22 / 525, 1 / 40],
        dtype=float,
    )
    # Corresponds to the optimum value of c_6 from [2]_.
    P = np.ascontiguousarray(
        [
            [
                1,
                -8048581381 / 2820520608,
                8663915743 / 2820520608,
                -12715105075 / 11282082432,
            ],
            [0, 0, 0, 0],
            [
                0,
                131558114200 / 32700410799,
                -68118460800 / 10900136933,
                87487479700 / 32700410799,
            ],
            [
                0,
                -1754552775 / 470086768,
                14199869525 / 1410260304,
                -10690763975 / 1880347072,
            ],
            [
                0,
                127303824393 / 49829197408,
                -318862633887 / 49829197408,
                701980252875 / 199316789632,
            ],
            [
                0,
                -282668133 / 205662961,
                2019193451 / 616988883,
                -1453857185 / 822651844,
            ],
            [0, 40617522 / 29380423, -110615467 / 29380423, 69997945 / 29380423],
        ],
        dtype=float,
    )

    _step_builder = step_builder

    @classmethod
    def _step_builder_args(cls):
        return cls.A, cls.B, cls.C, cls.E, cls.error_exponent



"""
    nbkode.testsuite.solver_fs
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test fixed step solvers for:
    - CaseInsensitiveDict

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import pytest

from nbkode.util import CaseInsensitiveDict


class TestCaseInsensitiveDict:

    # Adapted from https://github.com/psf/requests/blob/master/tests/test_structures.py

    @pytest.fixture(autouse=True)
    def setup(self):
        """CaseInsensitiveDict instance with "Accept" header."""
        self.case_insensitive_dict = CaseInsensitiveDict()
        self.case_insensitive_dict["Accept"] = "application/json"

    def test_list(self):
        assert list(self.case_insensitive_dict) == ["Accept"]

    possible_keys = pytest.mark.parametrize(
        "key", ("accept", "ACCEPT", "aCcEpT", "Accept")
    )

    @possible_keys
    def test_getitem(self, key):
        assert self.case_insensitive_dict[key] == "application/json"

    @possible_keys
    def test_delitem(self, key):
        del self.case_insensitive_dict[key]
        assert key not in self.case_insensitive_dict

    def test_lower_items(self):
        assert list(self.case_insensitive_dict.lower_items()) == [
            ("accept", "application/json")
        ]

    def test_repr(self):
        assert repr(self.case_insensitive_dict) == "{'Accept': 'application/json'}"

    def test_copy(self):
        copy = self.case_insensitive_dict.copy()
        assert copy is not self.case_insensitive_dict
        assert copy == self.case_insensitive_dict

    @pytest.mark.parametrize(
        "other, result",
        (({"AccePT": "application/json"}, True), ({}, False), (None, False)),
    )
    def test_instance_equality(self, other, result):
        assert (self.case_insensitive_dict == other) is result



"""
    nbkode.testsuite.solver
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test solvers methods

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np
import pytest

import nbkode

solvers = nbkode.get_solvers()


y0_1 = np.atleast_1d(1.0)
y0_2 = np.atleast_1d([1.0, 2.0])


def f1(t, x, k):
    return -k * x


@pytest.mark.parametrize("solver", solvers)
def test_f1_public_api(solver):
    # This is just running the public api, not checking correctness
    solver: nbkode.core.Solver = solver(f1, 0.0, y0_1, params=(0.01,))
    assert solver.step() == 1
    assert solver.step(solver.t) == 0
    assert solver.nsteps(5) == 5
    assert solver.nsteps(5, solver.t) == 0
    assert isinstance(solver.t, float)
    assert solver.y.shape == y0_1.shape
    assert solver.f.shape == y0_1.shape

    t, y = solver.move_to(solver.t + 1)
    assert isinstance(t, float)
    assert y.shape == y0_1.shape

    t, y = solver.run(solver.t + 1)
    assert y.shape == (len(t),) + y0_1.shape

    ts = np.linspace(solver.t, solver.t + 10, 20)
    t, y = solver.run(ts)
    assert np.allclose(ts, t)
    assert y.shape == (len(t),) + y0_1.shape



"""
    nbkode.adams_moulton
    ~~~~~~~~~~~~~~~~~~~~

    Methods of the Adams–Moulton family (implicit methods).
    - AdamsMoulton1
    - AdamsMoulton2
    - AdamsMoulton3
    - AdamsMoulton4
    - AdamsMoulton5

    See: https://en.wikipedia.org/wiki/Linear_multistep_method

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


import numpy as np

from .corefs import BFixedStepBaseSolver


class _AdamsMoulton(BFixedStepBaseSolver):

    GROUP = "Adams-Moulton"


class AdamsMoulton1(_AdamsMoulton):
    """The Adams–Bashforth method with ONE is equivalent to Euler"""

    COEFS = 1.0


class AdamsMoulton2(_AdamsMoulton):
    """The Adams–Bashforth method with TWO.

    ::

        u[n+1] = u[n] + h * (3/2 *f(u[n], t[n]) - 1/2 * f(u[n-1], t[n-1]))
    """

    COEFS = np.asarray([1, 1]) / 2


class AdamsMoulton3(_AdamsMoulton):
    """The Adams–Bashforth method with THREE.

    ::

        u[n+1] = u[n] + h /12 .*(23*f(u[n], t[n]) - 16*f(u[n-1], t[n-1])
                                + 5*f(u[n-2], t[n-2]))

    """

    COEFS = np.asarray([5, 8, -1]) / 12


class AdamsMoulton4(_AdamsMoulton):
    """The Adams–Bashforth method with FOUR.

    ::

        u[n+1] = u[n] + h / 24 .* (55.*f(u[n], t[n]) - 59*f(u[n-1], t[n-1]) +
                                   37*f(u[n-2], t[n-2]) - 9*f(u[n-3], t[n-3]))

    """

    COEFS = np.asarray([9, 19, -5, 1]) / 24


class AdamsMoulton5(_AdamsMoulton):
    """The Adams–Bashforth method with FIVE.

    ::

        u[n+1] = u[n] + h / 24 .* (1901.*f(u[n], t[n]) - 2774*f(u[n-1], t[n-1]) +
                                   2616*f(u[n-2], t[n-2]) - 1274*f(u[n-3], t[n-3]) +
                                   251 *f(u[n-4], t[n-4]) )

    """

    COEFS = np.asarray([251, 646, -264, 106, -19]) / 720



"""
    benchmarks.common
    ~~~~~~~~~~~~~~~~~

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from enum import IntEnum, auto


class NumbaStepModes(IntEnum):
    NUMBA_DISABLED = auto()
    EXTERNAL_LOOP = auto()
    INTERNAL_LOOP = auto()



"""
    benchmarks.against_scipy
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Comparisons using SciPy as a gold standard.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np
from scipy import integrate

import nbkode.dop853
from nbkode import runge_kutta
from nbkode.nbcompat import numba

from .common import NumbaStepModes

by_name = {
    "scipy": {
        "RungeKutta23": integrate.RK23,
        "RungeKutta45": integrate.RK45,
        "DOP853": integrate.DOP853,
    },
    "nbkode": {
        "RungeKutta23": runge_kutta.RungeKutta23,
        "RungeKutta45": runge_kutta.RungeKutta45,
        "DOP853": nbkode.dop853.DOP853,
    },
}

y0 = np.atleast_1d(1.0)


def create_f(package):
    if package == "nbkode":

        def f(t, y, k):
            return k * y

        args = 0.0, y0, -0.01
    else:
        k = -0.01

        def f(t, y):
            return k * y

        args = 0.0, y0

    return f, args


PACKAGES = tuple(by_name.keys())
INTEGRATORS = tuple(by_name["scipy"].keys())
BOOLEANS = (True, False)
NUMBA_MODES = tuple(NumbaStepModes.__members__.keys())


func = None


def define_func(package, jit_rhs_before):
    global func

    if jit_rhs_before:
        func, args = create_f(package)
        func = numba.njit()(func)
    else:
        func, args = create_f(package)

    # Test (and compile) func
    func(*args)


sol = None


def define_sol(package, integrator):
    global sol, func
    solver_cls = by_name[package][integrator]

    if package == "nbkode":
        sol = solver_cls(func, 0.0, y0, args=(-0.01,))
    else:
        sol = solver_cls(func, 0.0, y0, t_bound=10_000_000_000)


###############
# Instantiate
###############


def setup_time_f1_instantiate(package, integrator, jit_rhs_before):
    define_func(package, jit_rhs_before)


def time_f1_instantiate(package, integrator, jit_rhs_before):
    """Measures the time required to instantiate the solver"""
    define_sol(package, integrator)


time_f1_instantiate.setup = setup_time_f1_instantiate
time_f1_instantiate.params = (PACKAGES, INTEGRATORS, BOOLEANS)
time_f1_instantiate.param_names = ["package", "integrator", "jit_rhs_before"]


###############
# First Step
###############


def setup_time_f1_first_step(package, integrator):
    define_func(package, True)
    define_sol(package, integrator)


def time_f1_first_step(package, integrator):
    sol.step()


time_f1_first_step.setup = setup_time_f1_first_step
time_f1_first_step.params = (PACKAGES, INTEGRATORS)
time_f1_first_step.param_names = ["package", "integrator"]


###############
# Run 10k
###############


def setup_time_f1_run10k(package, integrator, other):
    if other == NumbaStepModes.INTERNAL_LOOP.name and package == "scipy":
        raise NotImplementedError

    define_func(package, True)
    define_sol(package, integrator)
    sol.step()
    if hasattr(sol, "nsteps"):
        sol.nsteps(1)


def time_f1_run10k(package, integrator, other):
    if other == NumbaStepModes.INTERNAL_LOOP.name:
        sol.nsteps(10_000)
    else:
        for n in range(10_000):
            sol.step()


time_f1_run10k.setup = setup_time_f1_run10k
time_f1_run10k.params = (
    PACKAGES,
    INTEGRATORS,
    (NumbaStepModes.INTERNAL_LOOP.name, NumbaStepModes.EXTERNAL_LOOP.name),
)
time_f1_run10k.param_names = ["package", "integrator", "other"]



"""
    nbkode.nb_to_import
    ~~~~~~~~~~~~~~~~~~~

    Select which numba to import based on NBKODE_NONUMBA environmental variable.

    We keep this here to allow importing within nbkode.nbcompat module
    and avoid circular imports.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import os

if os.environ.get("NBKODE_NONUMBA", 0):
    from . import numbasub as numba  # noqa: F401
else:
    import numba  # noqa: F401



"""
    nbkode.testsuite
    ~~~~~~~~~~~~~~~~

    numbakit-ode (nbkode) is a Python package to solve
    **ordinary differential equations (ODE)** that uses
    numba to compile code and therefore speed up calculations.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


def run():
    """Run all tests."""

    try:
        import pytest
    except ImportError:
        print("pytest not installed. Install it\n    pip install pytest")
        raise

    return pytest.main()



"""
    nbkode.euler
    ~~~~~~~~~~~~

    Methods of the Adams–Moulton family (implicit methods).
    - ForwardEuler (Euler)
    - BackwardEuler

    See: https://en.wikipedia.org/wiki/Linear_multistep_method

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .corefs import BFixedStepBaseSolver, FFixedStepBaseSolver


class ForwardEuler(FFixedStepBaseSolver):
    """The simple explicit (forward) Euler scheme

    y[n+1] = y[n] + f(t[n], y[n]) * h
    """

    GROUP = "Euler"

    COEFS = 1.0


Euler = ForwardEuler


class BackwardEuler(BFixedStepBaseSolver):
    """The simple explicit (forward) Euler scheme

    y[n+1] = y[n] + f(t[n+1], y[n+1]) * h
    """

    GROUP = "Euler"
    COEFS = 1.0



"""
    nbkode.nbcompat.numbasub
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Noop decorators disguised as numba.

    Copied from https://github.com/ptooley/numbasub

    See also: https://stackoverflow.com/questions/3888158

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import functools


def optional_arg_decorator(fn):
    @functools.wraps(fn)
    def wrapped_decorator(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            return fn(args[0])

        else:

            def real_decorator(decoratee):
                return fn(decoratee, *args, **kwargs)

            return real_decorator

    return wrapped_decorator


@optional_arg_decorator
def __noop(func, *args, **kwargs):
    return func


autojit = __noop
generated_jit = __noop
guvectorize = __noop
jit = __noop
jitclass = __noop
njit = __noop
vectorize = __noop

b1 = None
bool_ = None
boolean = None
byte = None
c16 = None
c8 = None
char = None
complex128 = None
complex64 = None
double = None
f4 = None
f8 = None
ffi = None
ffi_forced_object = None
float32 = None
float64 = None
float_ = None
i1 = None
i2 = None
i4 = None
i8 = None
int16 = None
int32 = None
int64 = None
int8 = None
int_ = None
intc = None
intp = None
long_ = None
longlong = None
none = None
short = None
u1 = None
u2 = None
u4 = None
u8 = None
uchar = None
uint = None
uint16 = None
uint32 = None
uint64 = None
uint8 = None
uintc = None
uintp = None
ulong = None
ulonglong = None
ushort = None
void = None



"""
    benchmarks.against_scipy
    ~~~~~~~~~~~~~~~~~~~~~~~~

    Comparisons of the different methods with and without numba.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import os

import numpy as np

from .common import NumbaStepModes

# Do not import this automatically as we want to be able to disable Numba
# and this requires not importing nbkode soon.
names = [
    "AdamsBashforth1",
    "AdamsBashforth2",
    "AdamsBashforth3",
    "AdamsBashforth4",
    "AdamsBashforth5",
    "AdamsMoulton1",
    "AdamsMoulton2",
    "AdamsMoulton3",
    "AdamsMoulton4",
    "AdamsMoulton5",
    "ForwardEuler",
    "BackwardEuler",
    "RungeKutta23",
    "RungeKutta45",
    "DOP853",
]

BOOLEANS = (True, False)
NUMBA_MODES = tuple(NumbaStepModes.__members__.keys())


y0 = np.atleast_1d(1.0)


func = None


def define_func(numba_enabled):
    global func
    if not numba_enabled:
        os.environ["NBKODE_NONUMBA"] = "1"

    # Leave the import inside to ensure proper NBKODE_NONUMBA
    import nbkode  # noqa: F401
    from nbkode.nbcompat import numba

    @numba.njit()
    def f(t, x, k):
        return k * x

    _ = f(0.0, y0, -0.01)
    func = f


sol = None


def define_sol(integrator):
    global sol

    # Leave the import inside to ensure proper NBKODE_NONUMBA
    import nbkode

    solvers = {solver.__name__: solver for solver in nbkode.get_solvers()}
    solver_cls = solvers[integrator]

    sol = solver_cls(func, 0.0, y0, args=(-0.01,))


###############
# Instantiate
###############


def setup_instantiate(integrator, numba_enabled):
    define_func(numba_enabled)


def time_instantiate(integrator, numba_enabled):
    return define_sol(integrator)


time_instantiate.setup = setup_instantiate
time_instantiate.params = (names, BOOLEANS)
time_instantiate.param_names = ("integrator", "numba_enabled")


###############
# First Step
###############


def setup_time_f1_first_step(integrator, numba_enabled):
    define_func(numba_enabled)
    define_sol(integrator)


def time_f1_first_step(integrator, numba_enabled):
    sol.step()


time_f1_first_step.setup = setup_time_f1_first_step
time_f1_first_step.params = (names, BOOLEANS)
time_f1_first_step.param_names = ("integrator", "numba_enabled")


###############
# Run 10k
###############


def setup_time_f1_run10k(integrator, other):
    numba_enabled = other != NumbaStepModes.NUMBA_DISABLED.name
    define_func(numba_enabled)
    define_sol(integrator)
    sol.step()
    if hasattr(sol, "nsteps"):
        sol.nsteps(1)


def time_f1_run10k(integrator, other):
    if other == NumbaStepModes.INTERNAL_LOOP.name:
        sol.nsteps(10_000)
    else:
        for n in range(10_000):
            sol.step()


time_f1_run10k.setup = setup_time_f1_run10k
time_f1_run10k.params = (names, NUMBA_MODES)
time_f1_run10k.param_names = ("integrator", "other")



"""
    nbkode
    ~~~~~~

    numbakit-ode (nbkode) is a Python package to solve
    **ordinary differential equations (ODE)** that uses
    numba to compile code and therefore speed up calculations.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .adams_bashforth import (
    AdamsBashforth1,
    AdamsBashforth2,
    AdamsBashforth3,
    AdamsBashforth4,
    AdamsBashforth5,
)
from .adams_moulton import (
    AdamsMoulton1,
    AdamsMoulton2,
    AdamsMoulton3,
    AdamsMoulton4,
    AdamsMoulton5,
)
from .core import get_groups, get_solvers
from .dop853 import DOP853
from .euler import BackwardEuler, Euler, ForwardEuler
from .runge_kutta import RungeKutta23, RungeKutta45

try:
    from importlib.metadata import version
except ImportError:
    # Backport for Python < 3.8
    from importlib_metadata import version

try:  # pragma: no cover
    __version__ = version("numbakit-ode")
except Exception:  # pragma: no cover
    # we seem to have a local copy not installed without setuptools
    # so the reported version will be unknown
    __version__ = "unknown"


def test():  # pragma: no cover
    """Run all tests.

    Returns
    -------
    unittest.TestResult
    """
    from .testsuite import run

    return run()



"""
    nbkode.corevs
    ~~~~~~~~~~~~~

    Definitions for Variable Step Methods.

    See: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from typing import Callable

import numpy as np
from scipy.integrate._ivp.common import (
    select_initial_step,
    validate_max_step,
    validate_tol,
)

from nbkode.nbcompat import numba

from .core import Solver

SAFETY = 0.9  # Multiply steps computed from asymptotic behaviour of errors by this.
MIN_FACTOR = 0.2  # Minimum allowed decrease in a step size.
MAX_FACTOR = 10  # Maximum allowed increase in a step size.


@numba.njit()
def rk_step(rhs, t, y, f, h, A, B, C, K):
    """Perform a single Runge-Kutta step.
    This function computes a prediction of an explicit Runge-Kutta method and
    also estimates the error of a less accurate method.
    Notation for Butcher tableau is as in [1]_.

    Parameters
    ----------
    fun : callable
        Right-hand side of the system.
    t : float
        Current time.
    y : ndarray, shape (n,)
        Current state.
    f : ndarray, shape (n,)
        Current value of the derivative, i.e., ``fun(x, y)``.
    h : float
        Step to use.
    A : ndarray, shape (n_stages, n_stages)
        Coefficients for combining previous RK stages to compute the next
        stage. For explicit methods the coefficients at and above the main
        diagonal are zeros.
    B : ndarray, shape (n_stages,)
        Coefficients for combining RK stages for computing the final
        prediction.
    C : ndarray, shape (n_stages,)
        Coefficients for incrementing time for consecutive RK stages.
        The value for the first stage is always zero.
    K : ndarray, shape (n_stages + 1, n)
        Storage array for putting RK stages here. Stages are stored in rows.
        The last row is a linear combination of the previous rows with
        coefficients

    Returns
    -------
    y_new : ndarray, shape (n,)
        Solution at t + h computed with a higher accuracy.
    f_new : ndarray, shape (n,)
        Derivative ``fun(t + h, y_new)``.

    References
    ----------
    .. [1] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
           Equations I: Nonstiff Problems", Sec. II.4.
    """
    K[0] = f
    for s in range(1, len(C)):
        a, c = A[s], C[s]
        dy = (K[:s].T @ a[:s]) * h
        K[s] = rhs(t + c * h, y + dy)

    y_new = y + h * (K[:-1].T @ B)
    f_new = rhs(t + h, y_new)
    K[-1] = f_new

    return y_new, f_new


class VariableStepRungeKutta(Solver):
    """Base class for explicit Runge-Kutta methods.

    Parameters
    ----------
    max_step : float, optional
        Maximum allowed step size. Default is np.inf, i.e., the step size is not
        bounded and determined solely by the solver.
    rtol, atol : float and array_like, optional
        Relative and absolute tolerances. The solver keeps the local error
        estimates less than ``atol + rtol * abs(y)``. Here, `rtol` controls a
        relative accuracy (number of correct digits). But if a component of `y`
        is approximately below `atol`, the error only needs to fall within
        the same `atol` threshold, and the number of correct digits is not
        guaranteed. If components of y have different scales, it might be
        beneficial to set different `atol` values for different components by
        passing array_like with shape (n,) for `atol`. Default values are
        1e-3 for `rtol` and 1e-6 for `atol`.

    See `Solver` for information about the other arguments.
    """

    FIXED_STEP = False

    C: np.ndarray
    A: np.ndarray
    B: np.ndarray
    E: np.ndarray
    P: np.ndarray
    order: int
    error_estimator_order: int
    n_stages: int
    error_exponent: float

    def __init_subclass__(cls, **kwargs):
        cls.error_exponent = -1 / (cls.error_estimator_order + 1)
        super().__init_subclass__(**kwargs)

    def __init__(
        self,
        rhs: Callable,
        t0: float,
        y0: np.ndarray,
        params: np.ndarray = None,
        *,
        rtol=1e-3,
        atol=1e-6,
        max_step=np.inf,
    ):
        super().__init__(rhs, t0, y0, params)

        self.max_step = validate_max_step(max_step)
        self.rtol, self.atol = validate_tol(rtol, atol, self.y.size)

        h = select_initial_step(
            self.rhs,
            self.t,
            self.y,
            self.f,
            1,
            self.error_estimator_order,
            self.rtol,
            self.atol,
        )

        # This is required to allow editing in place.
        self.h = np.atleast_1d(h)
        self.K = np.empty((self.n_stages + 1, self.y.size), dtype=float)

    def _step_extra_args(self):
        return self.h, self.K, self.rtol, self.atol, self.max_step



"""
    nbkode.core
    ~~~~~~~~~~~

    Definition for Solver base class.


    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from __future__ import annotations

from abc import ABC, ABCMeta, abstractmethod
from collections.abc import Collection
from numbers import Real
from typing import Callable, Tuple, Union

import numpy as np

from .nbcompat import numba
from .util import CaseInsensitiveDict


class MetaSolver(ABCMeta):
    def __repr__(cls):
        return f"<{cls.__name__}>"


class Solver(ABC, metaclass=MetaSolver):
    """Base class for all solvers

    Parameters
    ----------
    rhs : callable
        Right-hand side of the system. The calling signature is ``fun(t, y)``.
        Here ``t`` is a scalar, andthe ndarray ``y`` hasna shape (n,);
        then ``fun`` must return array_like with shape (n,).
    t0 : float
        Initial time.
    y0 : array_like, shape (n,)
        Initial state.
    params : array_like
        Extra arguments to be passed to the fun as ``fun(t, y, params)``

    Attributes
    ----------
    t : float
        Current time.
    y : ndarray
        Current state.
    f : ndarray
        last evaluation of the rhs.
    step_size : float
        Size of the last successful step. None if no steps were made yet.
    """

    SOLVERS = CaseInsensitiveDict()

    LEN_HISTORY: int = 2

    GROUP: str = None
    IMPLICIT: bool
    FIXED_STEP: bool

    #: Callable provided by the user
    #: The signature should be (t: float, y: ndarray)  -> ndarray
    #: or
    #: The signature should be (t: float, y: ndarray, p: ndarray)  -> ndarray
    rhs: Callable

    #: user rhs (same as rhs if it was originally jitted and with the right signature)
    user_rhs: Callable

    #: Last LEN_HISTORY times
    _ts: np.ndarray

    #: Last LEN_HISTORY states
    _ys: np.ndarray

    #: Last LEN_HISTORY evaluations of fun at (_t, _y)
    _ys: np.ndarray

    #: extra arguments for the user callable
    params: np.ndarray or None

    #: Function that build the _step function for a particular method.
    #: (*args) -> Callable
    _step_builder: Callable

    def __init__(
        self, rhs: Callable, t0: float, y0: np.ndarray, params: np.ndarray = None
    ):

        y0 = np.ascontiguousarray(y0)
        if params is not None:
            params = np.ascontiguousarray(params)

        self.user_rhs = rhs

        # TODO: is therse something more robust, such as a numba function
        if not hasattr(rhs, "inspect_llvm"):
            rhs = numba.njit()(rhs)

        # TODO: A better way to make it partial?
        if params is None:
            self.rhs = rhs
        else:

            @numba.njit()
            def _rhs(t, y):
                return rhs(t, y, params)

            self.rhs = _rhs

        f = self.rhs(t0, y0)
        self._fs = np.full((self.LEN_HISTORY, y0.size), f, dtype=float)
        self._ts = np.full(self.LEN_HISTORY, t0, dtype=float)
        self._ys = np.full((self.LEN_HISTORY, y0.size), y0, dtype=float)

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.LEN_HISTORY and cls.LEN_HISTORY < 1:
            raise ValueError(
                f"While defining {cls.__name__}, "
                f"LEN_HISTORY cannot be smaller than 1"
            )
        if cls.is_final_class():
            if cls.GROUP not in cls.SOLVERS:
                cls.SOLVERS[cls.GROUP] = []
            cls.SOLVERS[cls.GROUP].append(cls)
            cls._step = staticmethod(cls._step_builder(*cls._step_builder_args()))

    @classmethod
    @abstractmethod
    def _step_builder_args(cls):
        """Arguments provided to the _step_builder function."""

    @classmethod
    def is_final_class(cls):
        """True if the class represents a method and not a family/group of methods."""
        return cls.GROUP and not cls.__name__.startswith("_")

    @property
    def t(self):
        return self._ts[-1]

    @property
    def y(self):
        return self._ys[-1]

    @property
    def f(self):
        return self._fs[-1]

    def _check_time(self, t):
        if t < self.t:
            raise ValueError(f"Time {t} is smaller than current solver time t={self.t}")

    def step(self, t_bound=np.inf):
        """Advance simulation one time step.

        Parameters
        ----------
        t_bound : float, optional (default np.inf)
            The integration won’t continue beyond this value.
            In fixed step methods, the integration stops just before t_bound.
            In variable step methods, the integration stops at t_bound.

        Returns
        -------
        int
            number of steps given
        """
        return self._step(t_bound, *self._step_args(), *self._step_extra_args())

    def nsteps(self, steps: int, t_bound=np.inf):
        """Advance simulation multiple time step in a tight loop.

        Compiled equivalent of::

            cnt = 0
            for _ in range(steps):
                cnt += self.step(t_bound)
            return cnt

        Parameters
        ----------
        steps : int
            Number of steps to
        t_bound : float, optional (default np.inf)
            The integration won’t continue beyond this value.
            In fixed step methods, the integration stops just before t_bound.
            In variable step methods, the integration stops at t_bound.

        Returns
        -------
        int
            number of steps given
        """
        return self._nsteps(
            steps, t_bound, self._step, *self._step_args(), *self._step_extra_args()
        )

    def interpolate(self, t: float) -> float:
        """Interpolate solution at t.

        This only works for values within the recorded history of the solver.
        of the solver instance

        Parameters
        ----------
        t : float

        Raises
        ------
        ValueError
            if the time is outside the recorded history.
        """

        # TODO: make this work for array T

        if not (self._ts[0] <= t <= self._ts[-1]):
            raise ValueError(f"Time {t} to interpolate outside range")

        return self._interpolate(t, *self._step_args(), *self._step_extra_args())

    def move_to(self, t: float) -> Tuple[float, np.ndarray]:
        """Advance simulation up to t.

        Unlike `run`, this method does not output the intermediate steps.

        Parameters
        ----------
        t : float

        Returns
        -------
        float, np.ndarray
            time, state
        """
        self._check_time(t)
        return self._move_to(
            t, self._step, *self._step_args(), *self._step_extra_args()
        )

    def run(self, t: Union[Real, np.ndarray]) -> Tuple[np.ndarray, np.ndarray]:
        """Run solver.

        Parameters
        ----------
        t : float or array-like
            If t is a scalar, run freely up to t.
            If t is an array-like, return solution evaluated at t.

        Returns
        -------
        np.ndarray, np.ndarray
            time vector, state vector
        """
        if isinstance(t, Real):
            self._check_time(t)
            return self._run_free(
                t, self._step, *self._step_args(), *self._step_extra_args()
            )
        elif isinstance(t, Collection):
            t = np.asarray(t)
            self._check_time(t[0])
            return self._run_eval(
                t,
                self._step,
                self._interpolate,
                *self._step_args(),
                *self._step_extra_args(),
            )

    @staticmethod
    @abstractmethod
    def _step(t_bound, rhs, ts, ys, fs, *extra_args):
        """Numba-compiled."""
        raise NotImplementedError

    def _step_args(self):
        return self.rhs, self._ts, self._ys, self._fs

    def _step_extra_args(self) -> tuple:
        """Arguments to be passed to _step after t_bound."""
        return tuple()

    @staticmethod
    @numba.njit
    def _nsteps(steps, t_bound, step, rhs, ts, ys, fs, *extra_args):
        cnt = 0
        for _ in range(steps):
            tmp = step(t_bound, rhs, ts, ys, fs, *extra_args)
            if tmp == 0:
                return cnt
            cnt += tmp
        return cnt

    @staticmethod
    @numba.njit()
    def _interpolate(t_eval, rhs, ts, ys, fs, *extra_args):
        """Interpolate solution at t_eval"""
        y_out = np.empty(ys[0].shape)
        for ndx in range(len(y_out)):
            y_out[ndx] = np.interp(t_eval, ts, ys[:, ndx])

        return y_out

    @staticmethod
    @numba.njit
    def _move_to(t_end: float, step, rhs, ts, ys, fs, *extra_args):
        while ts[-1] < t_end:
            if not step(t_end, rhs, ts, ys, fs, *extra_args):
                break
        return ts[-1], ys[-1]

    @staticmethod
    @numba.njit
    def _run_free(
        t_end: float, step, rhs, ts, ys, fs, *extra_args
    ) -> tuple[np.ndarray, np.ndarray]:
        """Run freely upto t and return (t, y) as arrays."""
        t_out = [ts[-1]]
        y_out = [ys[-1]]

        while ts[-1] < t_end:
            if not step(t_end, rhs, ts, ys, fs, *extra_args):
                break
            t_out.append(ts[-1])
            y_out.append(ys[-1])

        out = np.empty((len(y_out), ys.shape[1]))
        for i, yi in enumerate(y_out):
            out[i] = yi

        return np.array(t_out), out

    @staticmethod
    @numba.njit
    def _run_eval(
        t_eval: np.ndarray, step, interpolate, rhs, ts, ys, fs, *extra_args
    ) -> tuple[np.ndarray, np.ndarray]:
        """Run up to t, evaluating y at given t and return (t, y) as arrays."""

        t_bound = t_eval[-1]
        y_eval = np.empty((t_eval.size, ys.shape[1]))

        for i, ti in enumerate(ts):
            while ts[-1] < ti:
                if not step(t_bound, rhs, ts, ys, fs, *extra_args):
                    break
            y_eval[i] = interpolate(ti, rhs, ts, ys, fs, *extra_args)

        return t_eval, y_eval


def check(solver, implicit=None, fixed_step=None):
    if implicit is not None:
        if solver.IMPLICIT is not implicit:
            return False
    if fixed_step is not None:
        if solver.FIXED_STEP is not fixed_step:
            return False
    return True


def get_solvers(*groups, implicit=None, fixed_step=None):
    """Get available solvers.

    Parameters
    ----------
    groups : str
        name of the group to filter
    implicit : bool
        if True, only implicit solvers will be returned.
    fixed_step : bool
        if True, only fixed step solvers will be returned.

    Returns
    -------
    tuple(Solver)
    """
    if not groups:
        groups = Solver.SOLVERS.keys()
    out = []
    for group in groups:
        try:
            out.extend(
                filter(
                    lambda solver: check(solver, implicit, fixed_step),
                    Solver.SOLVERS[group],
                )
            )
        except KeyError:
            m = tuple(Solver.SOLVERS.keys())
            raise KeyError(f"Group {group} not found. Valid values: {m}")
    return tuple(out)


def get_groups():
    """Get group names."""
    return tuple(sorted(Solver.SOLVERS.keys()))



"""
    nbkode.testsuite.test_against_scipy
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Comparisons using SciPy as a gold standard.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal
from scipy import integrate

import nbkode.dop853
from nbkode import runge_kutta

equivalents = [
    (runge_kutta.RungeKutta23, integrate.RK23),
    (runge_kutta.RungeKutta45, integrate.RK45),
    (nbkode.dop853.DOP853, integrate.DOP853),
]


def exponential1(t, x):
    return -0.01 * x


def exponential2(t, x):
    return np.asarray([-0.01, -0.05]) * x


y0_1 = np.atleast_1d(1.0)
y0_2 = np.atleast_1d([1.0, 2.0])


@pytest.mark.parametrize("nbkode_cls, scipy_cls", equivalents)
def test_exponential1(nbkode_cls, scipy_cls):
    nbkode_sol = nbkode_cls(exponential1, 0, y0_1)
    scipy_sol = scipy_cls(exponential1, 0, y0_1, t_bound=30)
    assert_equal(nbkode_sol.f, scipy_sol.f)
    assert nbkode_sol.h == scipy_sol.h_abs
    ndx = 0
    while True:
        ndx += 1
        nbkode_sol.step()
        scipy_sol.step()
        if scipy_sol.status != "running":
            break
        # We do not compare the last state as Scipy solvers are bound within step
        # and nbkode are not.
        msg = f"Step: {ndx}, Time: {scipy_sol.t}"
        assert nbkode_sol.t == scipy_sol.t, msg
        assert nbkode_sol.y == scipy_sol.y, msg
        assert nbkode_sol.f == scipy_sol.f, msg
        assert_allclose(nbkode_sol.h, scipy_sol.h_abs, err_msg=msg)
        assert_allclose(nbkode_sol.K, scipy_sol.K, err_msg=msg)


@pytest.mark.parametrize("nbkode_cls, scipy_cls", equivalents)
def test_exponential2(nbkode_cls, scipy_cls):
    nbkode_sol = nbkode_cls(exponential2, 0, y0_2)
    scipy_sol = scipy_cls(exponential2, 0, y0_2, t_bound=30)
    assert_equal(nbkode_sol.f, scipy_sol.f)
    assert nbkode_sol.h == scipy_sol.h_abs
    ndx = 0
    while True:
        ndx += 1
        nbkode_sol.step()
        scipy_sol.step()
        if scipy_sol.status != "running":
            break
        # We do not compare the last state as Scipy solvers are bound within step
        # and nbkode are not.
        msg = f"Step: {ndx}, Time: {scipy_sol.t}"
        assert nbkode_sol.t == scipy_sol.t, msg
        assert_equal(nbkode_sol.y, scipy_sol.y)
        assert_equal(nbkode_sol.f, scipy_sol.f)
        assert nbkode_sol.h == scipy_sol.h_abs, msg
        assert_allclose(nbkode_sol.K, scipy_sol.K, err_msg=msg)



"""
    nbkode.adams_bashforth
    ~~~~~~~~~~~~~~~~~~~~~~

    Methods of the Adams–Bashforth family (explicit methods).
    - AdamsBashforth1
    - AdamsBashforth2
    - AdamsBashforth3
    - AdamsBashforth4
    - AdamsBashforth5

    See: https://en.wikipedia.org/wiki/Linear_multistep_method

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np

from .corefs import FFixedStepBaseSolver


class _AdamsBashforth(FFixedStepBaseSolver):

    GROUP = "Adams-Bashforth"


class AdamsBashforth1(_AdamsBashforth):
    """The Adams–Bashforth method with ONE is equivalent to Euler"""

    COEFS = 1.0


class AdamsBashforth2(_AdamsBashforth):
    """The Adams–Bashforth method with TWO.

    ::

        u[n+1] = u[n] + h * (3/2 *f(u[n], t[n]) - 1/2 * f(u[n-1], t[n-1]))
    """

    COEFS = np.asarray([3, -1]) / 2


class AdamsBashforth3(_AdamsBashforth):
    """The Adams–Bashforth method with THREE.

    ::

        u[n+1] = u[n] + h /12 .*(23*f(u[n], t[n]) - 16*f(u[n-1], t[n-1])
                                + 5*f(u[n-2], t[n-2]))

    """

    COEFS = np.asarray([23, -16, 5]) / 12


class AdamsBashforth4(_AdamsBashforth):
    """The Adams–Bashforth method with FOUR.

    ::

        u[n+1] = u[n] + h / 24 .* (55.*f(u[n], t[n]) - 59*f(u[n-1], t[n-1]) +
                                   37*f(u[n-2], t[n-2]) - 9*f(u[n-3], t[n-3]))

    """

    COEFS = np.asarray([55, -59, 37, -9]) / 24


class AdamsBashforth5(_AdamsBashforth):
    """The Adams–Bashforth method with FIVE.

    ::

        u[n+1] = u[n] + h / 24 .* (1901.*f(u[n], t[n]) - 2774*f(u[n-1], t[n-1]) +
                                   2616*f(u[n-2], t[n-2]) - 1274*f(u[n-3], t[n-3]) +
                                   251 *f(u[n-4], t[n-4]) )

    """

    COEFS = np.asarray([1901, -2774, 2616, -1274, 251]) / 720



"""
    nbkode.nbcompat
    ~~~~~~~~~~~~~~~

    Numba compatible functions. Hopefully, some of these will be unnecessary
    in the future as Numba continues to grow.

    Current functions and their origin:
    - newton: scipy.optimize.newton
    - isclose: numpy.isclose
    - clip: numpy.clip

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from .common import clip, isclose
from .nb_to_import import numba
from .zeros import j_newton, newton, newton_hd

__all__ = ["clip", "isclose", "j_newton", "newton", "numba", "newton_hd"]



"""
    nbkode.testsuite.test_introspections
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Capabilities of the package provide information about
    the different solvers.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import nbkode


def test_get_solvers():
    assert len(nbkode.get_solvers()) == 15


def test_get_solvers_euler():
    assert len(nbkode.get_solvers("Euler")) == 2
    assert len(nbkode.get_solvers("euler")) == 2


def test_get_solvers_implicit():
    for solver in nbkode.get_solvers(implicit=True):
        assert solver.IMPLICIT is True
    for solver in nbkode.get_solvers(implicit=False):
        assert solver.IMPLICIT is False
    assert set(
        nbkode.get_solvers(implicit=False) + nbkode.get_solvers(implicit=True)
    ) == set(nbkode.get_solvers())


def test_get_solvers_fixed_step():
    for solver in nbkode.get_solvers(fixed_step=True):
        assert solver.FIXED_STEP is True
    for solver in nbkode.get_solvers(fixed_step=False):
        assert solver.FIXED_STEP is False
    assert set(
        nbkode.get_solvers(fixed_step=False) + nbkode.get_solvers(fixed_step=True)
    ) == set(nbkode.get_solvers())



"""
    nbkode.nbcompat.common
    ~~~~~~~~~~~~~~~~~~~~~~

    Common methods.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import numpy as np

from .nb_to_import import numba


@numba.njit()
def isclose(a, b, atol, rtol):
    return np.abs(a - b) <= (atol + rtol * np.abs(b))


@numba.njit
def clip(x, xmin, xmax):
    return min(max(x, xmin), xmax)



"""
    nbkode.nbcompat.newton
    ~~~~~~~~~~~~~~~~~~~~~~

    Newton methods.

    :copyright: 2020 by nbkode Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import enum
import operator
import warnings

import numpy as np
from scipy.optimize import zeros

from .common import isclose
from .nb_to_import import numba


class NewtonEnum(enum.IntEnum):
    OK = 0
    DERIVATIVE_WAS_ZERO = 1
    TOLERANCE_REACHED = 2
    NOT_CONVERGED = 3


@numba.njit
def _results_select(*args):
    """Select from a tuple of (root, funccalls, iterations, flag)"""
    return args


_ECONVERR = zeros._ECONVERR
_ECONVERGED = zeros._ECONVERGED


def newton(
    func,
    x0,
    fprime=None,
    args=(),
    tol=1.48e-8,
    maxiter=50,
    fprime2=None,
    x1=None,
    rtol=0.0,
    full_output=False,
    disp=True,
):
    """
    Find a zero of a real or complex function using the Newton-Raphson
    (or secant or Halley's) method.

    Find a zero of the function `func` given a nearby starting point `x0`.
    The Newton-Raphson method is used if the derivative `fprime` of `func`
    is provided, otherwise the secant method is used. If the second order
    derivative `fprime2` of `func` is also provided, then Halley's method is
    used.

    If `x0` is a sequence with more than one item, then `newton` returns an
    array, and `func` must be vectorized and return a sequence or array of the
    same shape as its first argument. If `fprime` or `fprime2` is given, then
    its return must also have the same shape.

    Parameters
    ----------
    func : callable
        The function whose zero is wanted. It must be a function of a
        single variable of the form ``f(x,a,b,c...)``, where ``a,b,c...``
        are extra arguments that can be passed in the `args` parameter.
    x0 : float, sequence, or ndarray
        An initial estimate of the zero that should be somewhere near the
        actual zero. If not scalar, then `func` must be vectorized and return
        a sequence or array of the same shape as its first argument.
    fprime : callable, optional
        The derivative of the function when available and convenient. If it
        is None (default), then the secant method is used.
    args : tuple, optional
        Extra arguments to be used in the function call.
    tol : float, optional
        The allowable error of the zero value. If `func` is complex-valued,
        a larger `tol` is recommended as both the real and imaginary parts
        of `x` contribute to ``|x - x0|``.
    maxiter : int, optional
        Maximum number of iterations.
    fprime2 : callable, optional
        The second order derivative of the function when available and
        convenient. If it is None (default), then the normal Newton-Raphson
        or the secant method is used. If it is not None, then Halley's method
        is used.
    x1 : float, optional
        Another estimate of the zero that should be somewhere near the
        actual zero. Used if `fprime` is not provided.
    rtol : float, optional
        Tolerance (relative) for termination.
    full_output : bool, optional
        If `full_output` is False (default), the root is returned.
        If True and `x0` is scalar, the return value is ``(x, r)``, where ``x``
        is the root and ``r`` is a `RootResults` object.
        If True and `x0` is non-scalar, the return value is ``(x, converged,
        zero_der)`` (see Returns section for details).
    disp : bool, optional
        If True, raise a RuntimeError if the algorithm didn't converge, with
        the error message containing the number of iterations and current
        function value. Otherwise, the convergence status is recorded in a
        `RootResults` return object.
        Ignored if `x0` is not scalar.
        *Note: this has little to do with displaying, however,
        the `disp` keyword cannot be renamed for backwards compatibility.*

    Returns
    -------
    root : float, sequence, or ndarray
        Estimated location where function is zero.
    r : `RootResults`, optional
        Present if ``full_output=True`` and `x0` is scalar.
        Object containing information about the convergence. In particular,
        ``r.converged`` is True if the routine converged.
    converged : ndarray of bool, optional
        Present if ``full_output=True`` and `x0` is non-scalar.
        For vector functions, indicates which elements converged successfully.
    zero_der : ndarray of bool, optional
        Present if ``full_output=True`` and `x0` is non-scalar.
        For vector functions, indicates which elements had a zero derivative.

    See Also
    --------
    brentq, brenth, ridder, bisect
    fsolve : find zeros in N dimensions.

    Notes
    -----
    The convergence rate of the Newton-Raphson method is quadratic,
    the Halley method is cubic, and the secant method is
    sub-quadratic. This means that if the function is well-behaved
    the actual error in the estimated zero after the nth iteration
    is approximately the square (cube for Halley) of the error
    after the (n-1)th step. However, the stopping criterion used
    here is the step size and there is no guarantee that a zero
    has been found. Consequently, the result should be verified.
    Safer algorithms are brentq, brenth, ridder, and bisect,
    but they all require that the root first be bracketed in an
    interval where the function changes sign. The brentq algorithm
    is recommended for general use in one dimensional problems
    when such an interval has been found.

    When `newton` is used with arrays, it is best suited for the following
    types of problems:

    * The initial guesses, `x0`, are all relatively the same distance from
      the roots.
    * Some or all of the extra arguments, `args`, are also arrays so that a
      class of similar problems can be solved together.
    * The size of the initial guesses, `x0`, is larger than O(100) elements.
      Otherwise, a naive loop may perform as well or better than a vector.

    Examples
    --------

    >>> from scipy import optimize
    >>> import matplotlib.pyplot as plt
    >>> def f(x):
    ...     return (x**3 - 1)  # only one real root at x = 1

    ``fprime`` is not provided, use the secant method:
    >>> root = optimize.newton(f, 1.5)
    >>> root
    1.0000000000000016
    >>> root = optimize.newton(f, 1.5, fprime2=lambda x: 6 * x)
    >>> root
    1.0000000000000016

    Only ``fprime`` is provided, use the Newton-Raphson method:
    >>> root = optimize.newton(f, 1.5, fprime=lambda x: 3 * x**2)
    >>> root
    1.0

    Both ``fprime2`` and ``fprime`` are provided, use Halley's method:
    >>> root = optimize.newton(f, 1.5, fprime=lambda x: 3 * x**2,
    ...                        fprime2=lambda x: 6 * x)
    >>> root
    1.0

    When we want to find zeros for a set of related starting values and/or
    function parameters, we can provide both of those as an array of inputs:
    >>> f = lambda x, a: x**3 - a
    >>> fder = lambda x, a: 3 * x**2
    >>> np.random.seed(4321)
    >>> x = np.random.randn(100)
    >>> a = np.arange(-50, 50)
    >>> vec_res = optimize.newton(f, x, fprime=fder, args=(a, ))

    The above is the equivalent of solving for each value in ``(x, a)``
    separately in a for-loop, just faster:
    >>> loop_res = [optimize.newton(f, x0, fprime=fder, args=(a0,))
    ...             for x0, a0 in zip(x, a)]
    >>> np.allclose(vec_res, loop_res)
    True

    Plot the results found for all values of ``a``:
    >>> analytical_result = np.sign(a) * np.abs(a)**(1/3)
    >>> fig = plt.figure()
    >>> ax = fig.add_subplot(111)
    >>> ax.plot(a, analytical_result, 'o')
    >>> ax.plot(a, vec_res, '.')
    >>> ax.set_xlabel('$a$')
    >>> ax.set_ylabel('$x$ where $f(x, a)=0$')
    >>> plt.show()

    """

    if tol <= 0:
        raise ValueError("tol too small (%g <= 0)" % tol)
    maxiter = operator.index(maxiter)
    if maxiter < 1:
        raise ValueError("maxiter must be greater than 0")
    if np.size(x0) > 1:
        return zeros._array_newton(
            func, x0, fprime, args, tol, maxiter, fprime2, full_output
        )

    (full_output, r), my_flag = _newton(
        func,
        x0,
        fprime,
        args,
        tol,
        maxiter,
        fprime2,
        x1,
        rtol,
        full_output,
    )

    if my_flag is not NewtonEnum.OK:

        if my_flag is NewtonEnum.DERIVATIVE_WAS_ZERO:
            msg = "Derivative was zero."
        elif my_flag is NewtonEnum.TOLERANCE_REACHED:
            msg = "Tolerance of ? reached."
        else:
            msg = ""

        if disp:
            p, funcalls, itr, flag = r
            if my_flag is NewtonEnum.NOT_CONVERGED:
                msg += "Failed to converge after %d iterations, value is %s." % (itr, p)
            else:
                msg += " Failed to converge after %d iterations, value is %s." % (
                    itr + 1,
                    p,
                )
            raise RuntimeError(msg)

        warnings.warn(msg, RuntimeWarning)

    return zeros._results_select(full_output, r)


def _newton(
    func,
    x0,
    fprime=None,
    args=(),
    tol=1.48e-8,
    maxiter=50,
    fprime2=None,
    x1=None,
    rtol=0.0,
    full_output=False,
):
    """Same signature as newton but:

    1. Has no `disp` argument
    2. Different output:
        full_output
        (p0, number of function calls, number of iterations, flag)
        NewtonEnum
    """

    my_flag = NewtonEnum.OK

    # Convert to float (don't use float(x0); this works also for complex x0)
    p0 = 1.0 * x0
    funcalls = 0
    if fprime is not None:
        # Newton-Raphson method
        for itr in range(maxiter):
            # first evaluate fval
            fval = func(p0, *args)
            funcalls += 1
            # If fval is 0, a root has been found, then terminate
            if fval == 0:
                return (
                    _results_select(full_output, (p0, funcalls, itr, _ECONVERGED)),
                    my_flag,
                )
            fder = fprime(p0, *args)
            funcalls += 1
            if fder == 0:
                my_flag = NewtonEnum.DERIVATIVE_WAS_ZERO
                # msg = "Derivative was zero."
                # if disp:
                #     msg += (
                #         " Failed to converge after %d iterations, value is %s."
                #         % (itr + 1, p0))
                #     raise RuntimeError(msg)
                # warnings.warn(msg, RuntimeWarning)
                return (
                    _results_select(full_output, (p0, funcalls, itr + 1, _ECONVERR)),
                    my_flag,
                )
            newton_step = fval / fder
            if fprime2:
                fder2 = fprime2(p0, *args)
                funcalls += 1
                # Halley's method:
                #   newton_step /= (1.0 - 0.5 * newton_step * fder2 / fder)
                # Only do it if denominator stays close enough to 1
                # Rationale: If 1-adj < 0, then Halley sends x in the
                # opposite direction to Newton. Doesn't happen if x is close
                # enough to root.
                adj = newton_step * fder2 / fder / 2
                if np.abs(adj) < 1:
                    newton_step /= 1.0 - adj
            p = p0 - newton_step
            if np.isclose(p, p0, rtol=rtol, atol=tol):
                return (
                    _results_select(full_output, (p, funcalls, itr + 1, _ECONVERGED)),
                    my_flag,
                )
            p0 = p
    else:
        # Secant method
        if x1 is not None:
            if x1 == x0:
                raise ValueError("x1 and x0 must be different")
            p1 = x1
        else:
            eps = 1e-4
            p1 = x0 * (1 + eps)
            p1 += eps if p1 >= 0 else -eps
        q0 = func(p0, *args)
        funcalls += 1
        q1 = func(p1, *args)
        funcalls += 1
        if np.abs(q1) < np.abs(q0):
            p0, p1, q0, q1 = p1, p0, q1, q0
        for itr in range(maxiter):
            if q1 == q0:
                if p1 != p0:
                    my_flag = NewtonEnum.TOLERANCE_REACHED
                    # msg = "Tolerance of %s reached." % (p1 - p0)
                    # if disp:
                    #     msg += (
                    #         " Failed to converge after %d iterations, value is %s."
                    #         % (itr + 1, p1))
                    #     raise RuntimeError(msg)
                    # warnings.warn(msg, RuntimeWarning)
                p = (p1 + p0) / 2.0
                return (
                    _results_select(full_output, (p, funcalls, itr + 1, _ECONVERGED)),
                    my_flag,
                )
            else:
                if np.abs(q1) > np.abs(q0):
                    p = (-q0 / q1 * p1 + p0) / (1 - q0 / q1)
                else:
                    p = (-q1 / q0 * p0 + p1) / (1 - q1 / q0)
            if isclose(p, p1, rtol=rtol, atol=tol):
                return (
                    _results_select(full_output, (p, funcalls, itr + 1, _ECONVERGED)),
                    my_flag,
                )
            p0, q0 = p1, q1
            p1 = p
            q1 = func(p1, *args)
            funcalls += 1

    # if disp:
    #     msg = ("Failed to converge after %d iterations, value is %s."
    #            % (itr + 1, p))
    #     raise RuntimeError(msg)

    return (
        _results_select(full_output, (p, funcalls, itr + 1, _ECONVERR)),
        NewtonEnum.NOT_CONVERGED,
    )


_j_newton = numba.njit()(_newton)


@numba.njit
def j_newton(
    func,
    x0,
    fprime=None,
    args=(),
    tol=1.48e-8,
    maxiter=50,
    fprime2=None,
    x1=None,
    rtol=0.0,
    full_output=False,
):
    """Same signature as newton but:

    1. Has no `disp` argument
    2. Different output:
        p0
    3. always raises RunTimeError if it does not converge.
    """

    (full_output, r), my_flag = _j_newton(
        func,
        x0,
        fprime,
        args,
        tol,
        maxiter,
        fprime2,
        x1,
        rtol,
        full_output,
    )
    p, funcalls, itr, flag = r
    if my_flag is not NewtonEnum.OK:
        raise RuntimeError
    return p


@numba.njit()
def jacobian(func, x, args=()):

    eps = 1e-10
    J = np.zeros((len(x), len(x)), dtype=np.float64)

    for i in range(len(x)):
        x1 = x.copy()
        x2 = x.copy()

        x1[i] += eps
        x2[i] -= eps

        f1 = func(x1, *args)
        f2 = func(x2, *args)

        J[:, i] = (f1 - f2) / (2 * eps)

    return J


@numba.njit()
def newton_hd_impl(
    func,
    y,
    args=(),
    atol=1.48e-8,
    rtol=0.0,
    maxiter=50,
):
    """
    Solve nonlinear system F=0 by Newton's method.

    J is the Jacobian of F. Both F and J must be functions of x.
    At input, x holds the start value. The iteration continues
    until ||F|| < eps.
    """
    value = func(y, *args)
    distance = np.linalg.norm(value, ord=2)  # l2 norm of vector
    iteration_counter = 0

    while not isclose(distance, 0, rtol=rtol, atol=atol):
        delta = np.linalg.solve(jacobian(func, y, args), -value)
        y = y + delta
        value = func(y, *args)
        distance = np.linalg.norm(value, ord=2)
        iteration_counter += 1

        if iteration_counter > maxiter:
            return y, NewtonEnum.NOT_CONVERGED

    return y, NewtonEnum.OK


@numba.njit()
def newton_hd(
    func,
    y0,
    args=(),
    atol=1.48e-8,
    rtol=0.0,
    maxiter=50,
):
    y, flag = newton_hd_impl(func, y0, args, atol, rtol, maxiter)
    if flag is not NewtonEnum.OK:
        raise RuntimeError
    return y
